/************************************************************
 **
 ** pf_op.c : perform operations on two constant values
 **
 ** Copyright (c) 2007, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include <assert.h>
#include "pf_expr.h"
#include "pf_op.h"

/* return n sign extended as necessary */
static variable_const_t pfile_op_val_sign_extend(
    variable_const_t     n,
    const variable_def_t def)
{
  variable_sz_t sz; /* *bit* size */

  assert((VARIABLE_DEF_TYPE_INTEGER == variable_def_type_get(def))
      || (VARIABLE_DEF_TYPE_BOOLEAN == variable_def_type_get(def)));

  sz = variable_def_sz_get(def);
  if (!variable_def_flag_test(def, VARIABLE_DEF_FLAG_BIT)) {
    sz *= 8;
  }
  if (VARIABLE_DEF_TYPE_BOOLEAN == variable_def_type_get(def)) {
    n = !!n;
  } else if (sz < 32) {
    if (variable_def_flag_test(def, VARIABLE_DEF_FLAG_SIGNED) 
        && (n & (1UL << (sz - 1)))) {
      n |= ~((1UL << sz) - 1);
    } else {
      n &= (1UL << sz) - 1;
    }
  }
  return n;
}

/* return the result if n is assigned */
static variable_const_t pfile_op_val_get(
    variable_const_t     n,
    const variable_def_t def)
{
  variable_sz_t sz;

  assert((VARIABLE_DEF_TYPE_INTEGER == variable_def_type_get(def))
      || (VARIABLE_DEF_TYPE_BOOLEAN == variable_def_type_get(def)));

  sz = variable_def_sz_get(def);
  if (!variable_def_flag_test(def, VARIABLE_DEF_FLAG_BIT)) {
    sz *= 8;
  }
  if (VARIABLE_DEF_TYPE_BOOLEAN == variable_def_type_get(def)) {
    n = !!n;
  } else {
    if (sz < 32) {
      n &= ((1 << sz) - 1);
    }
    if (variable_def_flag_test(def, VARIABLE_DEF_FLAG_BIT)
      && (1 == sz) 
      && variable_def_flag_test(def, VARIABLE_DEF_FLAG_SIGNED)) {
      n = -n;
    }
  }
  return pfile_op_val_sign_extend(n, def);
}

variable_const_t pfile_op_const_exec(pfile_t *pf,
    operator_t op,
    value_t    val1,
    value_t    val2)
{
  variable_def_t   def1;
  variable_def_t   def2;
  variable_def_t   rdef;
  variable_const_t n;
  variable_const_t n1;
  variable_const_t n2;
  variable_sz_t    sz;
  variable_const_t mask;
  variable_const_t sign_bit;
  boolean_t        result_is_signed;

  assert(value_is_const(val1) || value_is_pseudo_const(val1));
  assert(operator_is_unary(op) 
      || value_is_const(val2) 
      || value_is_pseudo_const(val2));

  def1 = value_def_get(val1);
  def2 = value_def_get(val2);
  rdef = pfile_variable_def_promotion_get(pf, op, val1, val2);

  n1               = pfile_op_val_get(value_const_get(val1), def1);
  n2               = (def2) 
                     ? pfile_op_val_get(value_const_get(val2), def2)
                     : 0;
  sz               = 8 * variable_def_sz_get(rdef);
  mask             = (sz < 32) ? (1UL << sz) - 1 : -1UL;
  sign_bit         = 1UL << 31;
  result_is_signed = variable_def_flag_test(
                        pfile_variable_def_promotion_get(pf, OPERATOR_NULL,
                        val1, val2), VARIABLE_DEF_FLAG_SIGNED);

  n = 0xdeadbeef; /* otherwise the compiler flags as maybe unassigned */
  switch (op) {
    case OPERATOR_NULL:
    case OPERATOR_REFERENCE:
    case OPERATOR_DOT:
    case OPERATOR_CAST:
    case OPERATOR_SUBSCRIPT:
    case OPERATOR_CT:
      assert(0);
      break;
    case OPERATOR_ADD:
      n = n1 + n2;
      break;
    case OPERATOR_SUB:
      n = n1 - n2;
      break;
    case OPERATOR_MUL:
      n = n1 * n2;
      break;
    case OPERATOR_DIV:
    case OPERATOR_MOD:
      if (!n2) {
        pfile_log(pf, PFILE_LOG_ERR, "divide by zero");
        n = (variable_const_t) ((value_is_same(val1, val2)) ? 1 : -1);
      } else {
        boolean_t sign;

        sign = BOOLEAN_FALSE;
        if (result_is_signed) {
          if (value_is_signed(val1) && (n1 & sign_bit)) {
            sign = BOOLEAN_TRUE;
            n1   = -n1;
          }
          if (value_is_signed(val2) && (n2 & sign_bit)) {
            sign ^= BOOLEAN_TRUE;
            n2    = -n2;
          }
        }
        n1 = n1 & mask;
        n2 = n2 & mask;
        n = (OPERATOR_DIV == op) ? n1 / n2 : n1 % n2;
        if (sign) {
          n = -n;
        }
      }
      break;
    case OPERATOR_LT:
    case OPERATOR_LE:
    case OPERATOR_EQ:
    case OPERATOR_NE:
    case OPERATOR_GE:
    case OPERATOR_GT:
      if (BOOLEAN_FALSE == result_is_signed) {
        variable_sz_t sz1;
        variable_sz_t sz2;

        sz1 = value_sz_get(val1);
        if (!value_is_bit(val1)) {
          sz1 *= 8;
        }
        sz2 = value_sz_get(val2);
        if (!value_is_bit(val2)) {
          sz2 *= 8;
        }
        sz = (variable_sz_t) ((sz1 < sz2) ? sz2 : sz1);
        if (sz < 8) {
          sz = 8;
        }
        mask = (sz < 32) ? (1UL << sz) - 1 : -1UL;
        n1 &= mask;
        n2 &= mask;
      }
      switch (op) {
        case OPERATOR_LT:
          n = (result_is_signed)
            ? ((long) n1 < (long) n2)
            : (n1 < n2);
          break;
        case OPERATOR_LE:
          n = (result_is_signed)
            ? ((long) n1 <= (long) n2)
            : (n1 <= n2);
          break;
        case OPERATOR_EQ:
          n = (n1 == n2);
          break;
        case OPERATOR_NE:
          n = (n1 != n2);
          break;
        case OPERATOR_GE:
          n = (result_is_signed)
            ? ((long) n1 >= (long) n2)
            : (n1 >= n2);
          break;
        case OPERATOR_GT:
          n = (result_is_signed)
            ? ((long) n1 > (long) n2)
            : (n1 > n2);
          break;
        default:
          break; /* cannot get here */
      }
      break;
    case OPERATOR_ANDL:
      n = n1 && n2;
      break;
    case OPERATOR_ORL:
      n = n1 || n2;
      break;
    case OPERATOR_NOTL:
      n = !n1;
      break;
    case OPERATOR_ANDB:
      n = n1 & n2;
      break;
    case OPERATOR_ORB:
      n = n1 | n2;
      break;
    case OPERATOR_XORB:
      n = n1 ^ n2;
      break;
    case OPERATOR_CMPB:
      if (value_is_boolean(val1)) {
        n = (variable_const_t) ((n1) ? 0 : -1);
      } else {
        n = ~n1;
      }
      break;
    case OPERATOR_ASSIGN:
      n = n1;
      break;
    case OPERATOR_NEG:
      n = -n1;
      break;
    case OPERATOR_INCR:
      n = n1 + 1;
      break;
    case OPERATOR_DECR:
      n = n1 - 1;
      break;
    case OPERATOR_SHIFT_LEFT:
      n = n1 << n2;
      break;
    case OPERATOR_SHIFT_RIGHT:
    case OPERATOR_SHIFT_RIGHT_ARITHMETIC:
      if (value_is_signed(val1)
          && (value_byte_sz_get(val1) < variable_def_sz_get(rdef))) {
        sign_bit = variable_def_sz_get(def1);
        if (!variable_def_flag_test(def1, VARIABLE_DEF_FLAG_BIT)) {
          sign_bit *= 8;
        }
      } else {
        sign_bit = variable_def_sz_get(rdef);
        if (!variable_def_flag_test(rdef, VARIABLE_DEF_FLAG_BIT)) {
          sign_bit *= 8;
        }
      }
      sign_bit = 1UL << (sign_bit - 1);
      if (n2 >= sz) {
        n = (variable_const_t) (((OPERATOR_SHIFT_RIGHT_ARITHMETIC == op)
            && (n1 & sign_bit))
          ? -1
          : 0);
      } else {
        n = (n1 & mask) >> n2;
        if ((OPERATOR_SHIFT_RIGHT_ARITHMETIC == op)
            && (sz - n2 < 32)
            && (n1 & sign_bit)) {
          n |= ~((1UL << (sz - n2)) - 1);
        }
      }
      n &= mask;
      break;
    case OPERATOR_LOGICAL:
      n = !!n1;
      break;
  }
  return pfile_op_val_get(n, rdef);
}

